function LootSquid:AddEvent(event, func)
    self.eventHandlers[event] = func
end

function LootSquid:RegisterEvents()

    self:LogDebug("RegisterEvents()")

    -- hook chat events
    self.eventHandlers["CHAT_MSG_BN_WHISPER"] = self.OnChatMessage
    self.eventHandlers["CHAT_MSG_RAID"] = self.OnChatMessage
    self.eventHandlers["CHAT_MSG_WHISPER"] = self.OnChatMessage
    --self.eventHandlers["CHAT_MSG_GUILD"]         = self.OnChatMessage
    --self.eventHandlers["CHAT_MSG_PARTY"]         = self.OnChatMessage
    self.eventHandlers["LOOT_OPENED"] = self.OnLootOpened
    self.eventHandlers["LOOT_CLOSED"] = self.OnLootClosed
    self.eventHandlers["LOOT_SLOT_CHANGED"] = self.OnLootSlotChanged
    self.eventHandlers["LOOT_SLOT_CLEARED"] = self.OnLootSlotCleared

    -- register all event handlers on the main frame
    for k in pairs(self.eventHandlers) do
        self.frames.mainFrame:RegisterEvent(k)
    end

    -- add on OnEvent handler on the main frame
    self.frames.mainFrame:SetScript("OnEvent",
        function(mainFrame, event, ...)
        --self:LogInfo("OnEvent: " .. event)
            self.eventHandlers[event](self, ...)
        end)
end

function LootSquid:IsPlayerMasterLooter()
    local master, partyMaster, raidMaster = GetLootMethod()
    local playerName = UnitName("player")

    local raidName = "raid" .. tostring(raidMaster)

    if master and raidMaster and playerName == UnitName(raidName) then
        self:LogDebug("IsPlayerMasterLooter: " .. tostring(true))
        return true
    else
        self:LogDebug("IsPlayerMasterLooter: " .. tostring(false))
        return false
    end
end

function LootSquid:OnLootSlotChanged(a, b, c, d, e, f, g, h)
    self:LogDebug("OnLootSlotChanged()")
    --    LootSquid:LogInfo("OnLootSlotChanged: "
    --            ..tostring(a)..", "..tostring(b)..", "..tostring(c)..", "..tostring(d)..", "
    --            ..tostring(e)..", "..tostring(f)..", "..tostring(g)..", "..tostring(h)..", ")
end

--[[
-- Fires when the contents of a loot slot are removed.
-- Can fire due to the player looting the slot's contents
-- or due to them being taken by another group member.
 ]]
function LootSquid:OnLootSlotCleared(event, slotId)
    self:LogDebug("OnLootSlotCleared. slotId: " .. tostring(slotId))
    -- slotId: The numeric id of the slot that was looted.
    LootSquid:UpdateLootTable(false)
end

function LootSquid:OnChatMessage(message, sender, language, channelString, target, flags,
unknown, channelNumber, channelName, unknown, counter, guid)

    if LootSquid.debug.bypassSafeCalls then
        LootSquid:HandleOnChatMessage(message, sender)
    else
        local success, error = pcall(LootSquid.HandleOnChatMessage, nil, message, sender)
        if not success then
            LootSquid:LogError("OnChatMessage failed: " .. tostring(error))
        end
    end
end

function LootSquid:HandleOnChatMessage(message, sender)
    LootSquid:LogDebug("OnChatMessage. Sender: " .. tostring(sender) .. ", message: " .. tostring(message))

    -- do not process chat message unless master looter
    if not (LootSquid.debug.alwaysParseChat or LootSquid:IsPlayerMasterLooter()) then
        return
    end

    local a, b, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, Name =
    string.find(message,
        "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*):?(%-?%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

    local itemLink2
    --    local itemTexture2
    --    local itemEquipLoc2
    --    local itemLevel2
    --    local itemQuality2

    if Id then
        local itemName, itemLink, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType,
        itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(Id)

        if itemName == nil then
            LootSquid:LogInfo("No local cache for item: " .. tostring(Id) .. ":" .. tostring(Name))
        end

        itemLink2 = itemLink
        --        itemTexture2 = itemTexture
        --        itemEquipLoc2 = itemEquipLoc
        --        itemLevel2 = itemLevel
        --        itemQuality2 = itemQuality
    end

    --local itemString = string.match(message, "item[%-?%d:]+")
    --    print(Id)
    --    print(LinkLvl)
    --    print(Name)

    -- if loot dist in session...

    LootSquid:LogDebug("OnChatMessage. Sender: " .. tostring(sender) .. ", message: " .. tostring(message) .. ". Parsing done.")

    local isNeed = LootSquid:strStartsWith(string.lower(message), "!need")
    local isGreed = LootSquid:strStartsWith(string.lower(message), "!greed")
    local isPass = LootSquid:strStartsWith(string.lower(message), "!pass")

    if (_G.LootSquid.session.currentLootSlot < 1 and (isNeed or isGreed)) then
        SendChatMessage("LootSquid: The loot master does not currently have an active loot session",
            "WHISPER",
            nil,
            sender)
        return
    end


    if (isNeed) then

        _G.LootSquid.session.playersNeeding[sender] = { isNeeding = true, currentItemLink = itemLink2 }
        _G.LootSquid.session.playersGreeding[sender] = nil
        LootSquid:UpdateRequestFrame()

        SendChatMessage("LootSquid: Your need request have been received. Please wait.",
            "WHISPER",
            nil,
            sender)
    end

    if (isGreed) then

        _G.LootSquid.session.playersGreeding[sender] = { isGreeding = true, currentItemLink = itemLink2 }
        _G.LootSquid.session.playersNeeding[sender] = nil
        LootSquid:UpdateRequestFrame()

        SendChatMessage("LootSquid: Your greed request have been received. Please wait.",
            "WHISPER",
            nil,
            sender)
    end

    if (isPass) then

        _G.LootSquid.session.playersGreeding[sender] = nil
        _G.LootSquid.session.playersNeeding[sender] = nil
        LootSquid:UpdateRequestFrame()

        SendChatMessage("LootSquid: Your pass request have been received. You have been removed from concideration.",
            "WHISPER",
            nil,
            sender)
    end
end

function LootSquid:UpdateLootTable(doOpenGui)
    for k in pairs(_G.LootSquid.session.loot) do
        _G.LootSquid.session.loot[k] = nil
    end

    local lootThreshold = GetLootThreshold()


    -- if player is the master looter
    if (_G.LootSquid.debug.alwaysParseLoot == true
            or (LootSquid:IsPlayerMasterLooter())) then

        local lootSlots = GetNumLootItems()
        local lootSlotsAtleastThreshold = 0

        _G.LootSquid.session.lootSlots = lootSlots

        for lootSlot = 1, lootSlots do

            --[[
            -- The lootSlot will remain concistent for items as long as the loot box remains open.
            -- If the loot box is closed after some items have been looted, lootSlot will have changed
            -- for items which originally was placed after a looted item.
            -- ]]
            local itemTexture, itemName, itemQuantity, itemQuality, itemLocked = GetLootSlotInfo(lootSlot)

            if itemQuality and itemQuality >= lootThreshold then
                lootSlotsAtleastThreshold = lootSlotsAtleastThreshold + 1
            end

            local itemLootType = GetLootSlotType(lootSlot)


            --            self:LogInfo("GetLootSlotInfo "
            --                    .. tostring(lootSlot) .. ", "
            --                    .. tostring(itemName) .. ", "
            --                    .. tostring(itemQuantity) .. ", "
            --                    .. tostring(itemQuality) .. ", "
            --                    .. tostring(itemLocked) .. ", "
            --                    .. tostring(itemLootType)
            --            )

            local itemLink = GetLootSlotLink(lootSlot)

            _G.LootSquid.session.loot[lootSlot] = {
                itemName = itemName,
                itemQuantity = itemQuantity,
                itemQuality = itemQuality,
                itemLocked = itemLocked,
                itemLootType = itemLootType,
                itemLink = itemLink,
                candidates = {}
            }

            for candidateIndex = 1, 40 do

                local candidateName = GetMasterLootCandidate(lootSlot, candidateIndex)
                if candidateName then

                    _G.LootSquid.session.loot[lootSlot].candidates[candidateName] = {
                        candidateIndex = candidateIndex,
                    }

                else
                    --LootSquid:LogInfo("no candidate name")
                end
            end
        end

        if (
                (_G.LootSquid.debug.alwaysParseLoot and doOpenGui)
                or
                (lootSlotsAtleastThreshold > 0
                        and doOpenGui
                        and lootSlots > 0
                --and (not InCombatLockdown() or IsShiftKeyDown())
                )
            ) then
            self:OpenLootGui()
        else
            self:OnUpdateLootTable()
        end
    end
end

function LootSquid:OnLootOpened(autoLoot)
    self:LogDebug("OnLootOpened. autoLoot: " .. tostring(autoLoot))
    self:UpdateLootTable(true)
end

function LootSquid:OnLootClosed()
end
